import pytest
import litemapy.storage as storage
import math

TEST_VALUES = 0, 0, 0, 12, 13, 0, 4, 0, 2, 4, 1, 3, 3, 7, 65, 9


def test_write_read_to_litematica_bit_array():
    nbits = math.ceil(math.log(max(TEST_VALUES), 2)) + 1
    arr = storage.LitematicaBitArray(len(TEST_VALUES), nbits)
    for i, e in enumerate(TEST_VALUES):
        arr[i] = e
    for i, e in enumerate(TEST_VALUES):
        print(i, e)
        assert e == arr[i]


def test_invalid_access_to_litematica_bit_array_raises_exceptions():
    arr = storage.LitematicaBitArray(10, 4)
    with pytest.raises(IndexError):
        arr[-1] = 0
    with pytest.raises(IndexError):
        arr[10] = 0
    with pytest.raises(ValueError):
        arr[0] = -1
    with pytest.raises(ValueError):
        arr[0] = 16


def test_litematica_bit_array_in():
    nbits = math.ceil(math.log(max(TEST_VALUES), 2)) + 1
    array = storage.LitematicaBitArray(len(TEST_VALUES), nbits)
    for i, e in enumerate(TEST_VALUES):
        array[i] = e
    assert 13 in array
    assert 15 not in array


def test_basic_set_get():
    def discriminator(k, v):
        print("Discriminating ", type(k), k, "=>", type(v), v)
        return v >= 0, "Need pos"

    discriminating_dictionary = storage.DiscriminatingDictionary(discriminator)
    discriminating_dictionary["0"] = 0
    assert discriminating_dictionary["0"] == 0
    assert "0" in discriminating_dictionary
    assert "0" in discriminating_dictionary.keys()
    assert 0 in discriminating_dictionary.values()
    assert discriminating_dictionary.get("1") is None
    with pytest.raises(storage.DiscriminationError):
        discriminating_dictionary['-1'] = -1
    other_dictionary = {"1": 1, "2": 2}
    discriminating_dictionary.update(other_dictionary)
    assert "1" in discriminating_dictionary
    assert "2" in discriminating_dictionary
    with pytest.raises(storage.DiscriminationError):
        discriminating_dictionary.update({"-1": -1})
    other_dictionary = {"1": 1, "2": 2}
    discriminating_dictionary = storage.DiscriminatingDictionary(lambda k, v: (v >= 0, "Need pos"), other_dictionary)
    assert "1" in discriminating_dictionary
    assert "2" in discriminating_dictionary
    discriminating_dictionary = storage.DiscriminatingDictionary(lambda k, v: (v >= 0, "Need pos"), a=1, b=2)
    assert "a" in discriminating_dictionary
    assert "b" in discriminating_dictionary


def test_discriminating_dictionary_onadd():
    class Counter:

        def __init__(self):
            self.counter = 0

        def on_add(self, k, v):
            self.counter += v

    c = Counter()
    dictionary = storage.DiscriminatingDictionary(
        lambda k, v: (v >= 0, "Need pos"),
        onadd=c.on_add,
        x=10
    )
    dictionary["a"] = 1
    assert c.counter == 1
    dictionary.update({"b": 2, "c": 3})
    assert c.counter == 6
    dictionary.setdefault("d", 4)
    assert c.counter == 10


def test_discriminating_dictionary_onremove():
    class Counter:
        def __init__(self):
            self.counter = 0

        def on_remove(self, k, v):
            self.counter += v

    c = Counter()
    dictionary = storage.DiscriminatingDictionary(
        lambda k, v: (v >= 0, "Need pos"),
        onremove=c.on_remove,
        a=1, b=2, c=3, d=4, x=10
    )
    del dictionary["a"]
    assert c.counter == 1
    dictionary.pop("b")
    assert c.counter == 3
    dictionary.pop("c")
    assert c.counter == 6
    dictionary.pop("d")
    assert c.counter == 10
    dictionary.popitem()
    assert c.counter == 20
    c = Counter()
    dictionary = storage.DiscriminatingDictionary(
        lambda k, v: (v >= 0, "Need pos"),
        onremove=c.on_remove,
        a=1, b=2, c=3, d=4, x=10
    )
    dictionary.clear()
    assert c.counter == 20


def test_discriminating_dictionary_onadd_onremove():
    class Counter:
        def __init__(self):
            self.added = 0
            self.removed = 0

        def on_remove(self, k, v):
            self.removed += v

        def on_add(self, k, v):
            self.added += v

    c = Counter()
    dictionary = storage.DiscriminatingDictionary(
        lambda k, v: (v >= 0, "Need pos"),
        onadd=c.on_add,
        onremove=c.on_remove,
        a=1, b=2, c=3, d=4, x=10
    )
    dictionary["c"] = 7
    assert c.added == 7
    assert c.removed == 3
    dictionary.update({"x": 100, "d": 500, "y": 200})
    assert c.added == 807
    assert c.removed == 17



from litemapy.deprecation import deprecated_name


def test_deprecated_name():

    class A:

        def __init__(self):
            self.g = None

        @deprecated_name("c")
        def b(self, arg, kwarg="default"):
            return arg, kwarg

        @deprecated_name("e")
        @staticmethod
        def d(arg, kwarg="default"):
            return arg, kwarg

        @property
        @deprecated_name("h")
        def prop(self):
            return self.g

        @prop.setter
        def prop(self, val):
            self.g = val

    a = A()
    assert a.b("hello", kwarg="world") == ("hello", "world")
    assert a.c("hello", kwarg="world") == ("hello", "world")
    assert a.d("lorem", kwarg="ipsum") == ("lorem", "ipsum")
    assert a.e("lorem", kwarg="ipsum") == ("lorem", "ipsum")
    assert A.d("lorem", kwarg="ipsum") == ("lorem", "ipsum")



import pytest

from litemapy import BlockState
from litemapy.minecraft import is_valid_identifier
from litemapy.minecraft import InvalidIdentifier
from litemapy.schematic import AIR


def test_blockstate_initialization():
    prop = {"test1": "testval", "test2": "testval2"}
    b = BlockState("minecraft:stone", **prop)
    assert len(prop) == len(b)
    for k, v in prop.items():
        assert b[k] == v


def test_cannot_create_blockstate_with_invalid_id():
    ids = (
        "",
        "minecraft stone",
        "stone",
        "minecraft:stone[property=value]",
    )
    for id_ in ids:
        with pytest.raises(InvalidIdentifier):
            BlockState(id_, prop="val")
        with pytest.raises(InvalidIdentifier):
            AIR.with_id(id_)


def test_blockstate_nbt_is_identity():
    prop = {"test1": "testval", "test2": "testval2"}
    blockstate_1 = BlockState("minecraft:stone", **prop)
    nbt = blockstate_1.to_nbt()
    blockstate_2 = BlockState.from_nbt(nbt)
    assert blockstate_1 == blockstate_2


def test_blockstate_with_properties():
    prop = {"test1": "testval1", "test2": "testval2"}
    blockstate_1 = BlockState("minecraft:stone", **prop)
    blockstate_2 = blockstate_1.with_properties(test3="testval3", test4="testval4")
    assert blockstate_2.to_block_state_identifier() == "minecraft:stone[test1=testval1,test2=testval2,test3=testval3,test4=testval4]"

    blockstate_3 = blockstate_2.with_properties(test4=None)
    assert blockstate_3.to_block_state_identifier() == "minecraft:stone[test1=testval1,test2=testval2,test3=testval3]"


def test_blockstate_is_hashable():
    state1 = BlockState("minecraft:air")
    state2 = BlockState("minecraft:air")
    assert state1 == state2
    assert hash(state1) == hash(state2)
    assert hash(state1) == hash(state1)


def test_is_valid_identifier():
    assert is_valid_identifier("minecraft:air")
    assert is_valid_identifier("minecraft:stone_cutter")
    assert is_valid_identifier("terramap:path/is/allowed_slashes.png")
    assert is_valid_identifier("weird.mod-id:both_are_allowed-dashes-and_underscores.and.dots")

    assert not is_valid_identifier("")
    assert not is_valid_identifier(" ")
    assert not is_valid_identifier("minecraft:minecraft:stone")
    assert not is_valid_identifier("minecraft:oak_stairs[facing=north]")
    assert not is_valid_identifier("minecraft")



LITEMATIC_VERSION = 6
LITEMATIC_SUBVERSION = 1
SPONGE_VERSION = 2  # Sponge format version used by WorldEdit (https://github.com/SpongePowered/Schematic-Specification)
MC_DATA_VERSION = 2975  # Minecraft 1.18.2 (https://minecraft.wiki/w/Data_version)
DEFAULT_NAME = "Unnamed"  # Default name given to schematics and regions if unspecified
LITEMAPY_NAME = "Litemapy"  # Used to identify schematic created with Litemapy in metadata
LITEMAPY_VERSION = "0.10.0b0"  # Unique version string



from math import ceil
import nbtlib.tag
from nbtlib import LongArray
from typing import Generator, Callable, Any, Optional


class LitematicaBitArray:
    size: int
    nbit: int
    array: list[int]
    __mask: int

    def __init__(self, size: int, nbits: int) -> None:
        self.size = size
        self.nbits = nbits
        s = ceil(nbits * size / 64)
        self.array = [0] * s
        self.__mask = (1 << nbits) - 1  # nbits bits set to 1

    @staticmethod
    def from_nbt_long_array(arr: LongArray, size: int, nbits: int) -> 'LitematicaBitArray':
        # TODO Test loading and validating from an external source
        expected_len = ceil(size * nbits / 64)
        if expected_len != len(arr):
            raise ValueError(
                "long array length does not match bit array size and nbits, expected {}, not {}".format(
                    expected_len, len(arr)
                )
            )
        r = LitematicaBitArray(size, nbits)
        m = (1 << 64) - 1
        r.array = [int(i) & m for i in arr]  # Remove the infinite trailing 1s of negative numbers
        return r

    def _to_long_list(self) -> list[int]:
        list_of_longs = []
        m1 = 1 << 63
        m2 = (1 << 64) - 1
        for i in self.array:
            if i & m1 > 0:
                i |= ~m2  # Add the potential infinite 1 prefix for negative numbers
            list_of_longs.append(i)
        return list_of_longs

    def _to_nbt_long_array(self) -> LongArray:
        return nbtlib.tag.LongArray(self._to_long_list())

    def __getitem__(self, index) -> int:
        if not 0 <= index < len(self):
            raise IndexError("Invalid index {}".format(index))
        start_offset = index * self.nbits
        start_arr_index = start_offset >> 6
        end_arr_index = ((index + 1) * self.nbits - 1) >> 6
        start_bit_offset = start_offset & 0x3F

        if start_arr_index == end_arr_index:
            return self.array[start_arr_index] >> start_bit_offset & self.__mask
        else:
            end_offset = 64 - start_bit_offset
            val = self.array[start_arr_index] >> start_bit_offset | self.array[end_arr_index] << end_offset
            return val & self.__mask

    def __setitem__(self, index: int, value: int) -> None:
        if not 0 <= index < len(self):
            raise IndexError("Invalid index {}".format(index))
        if not 0 <= value <= self.__mask:
            raise ValueError("Invalid value {}, maximum value is {}".format(value, self.__mask))
        start_offset = index * self.nbits
        start_arr_index = start_offset >> 6
        end_arr_index = ((index + 1) * self.nbits - 1) >> 6
        start_bit_offset = start_offset & 0x3F
        m = (1 << 64) - 1
        zeroed = self.array[start_arr_index] & ~(self.__mask << start_bit_offset)
        updated = zeroed | (value & self.__mask) << start_bit_offset
        self.array[start_arr_index] = updated & m

        if start_arr_index != end_arr_index:
            end_offset = 64 - start_bit_offset
            j1 = self.nbits - end_offset
            self.array[end_arr_index] = (self.array[end_arr_index] >> j1 << j1 | (
                    value & self.__mask) >> end_offset) & m

    def __len__(self) -> int:
        return self.size

    def __iter__(self) -> Generator[int, None, None]:
        for i in range(len(self)):
            yield self[i]

    def __reversed__(self) -> 'LitematicaBitArray':
        arr = LitematicaBitArray(self.size, self.nbits)
        for i in range(len(self)):
            arr[i] = self[len(self) - i - 1]
        return arr

    def __contains__(self, value: int) -> bool:
        for v in self:
            if v == value:
                return True
        return False


ValidatorFunction = Callable[[Any, Any], tuple[bool, str]]
ReactionFunction = Callable[[Any, Any], None]


class DiscriminatingDictionary(dict):
    validator: ValidatorFunction
    on_add: Optional[ReactionFunction]
    on_remove: Optional[ReactionFunction]

    def __init__(self, validator: ValidatorFunction, *args, onadd: Optional[ReactionFunction] = None,
                 onremove: Optional[ReactionFunction] = None, **options) -> None:
        """
        :params validator:  a function that takes as argument a key and an item and returns a tuple (canstore, msg)
                            canstore must be a boolean, True if the item is accepted, and False otherwise
                            if canstore is false, msg will be used as the error message
        :param onadd:       an optional function that gets called when an item is added to the dictionary,
                            with the key and item as arguments.
                            onadd is not called for the values that might have been passed in the constructor.
        """
        # TODO Handle iterators in constructor
        self.validator = validator
        self.on_add = onadd
        self.on_remove = onremove
        if len(args) == 1 and isinstance(args[0], dict):
            for key, item in args[0].items():
                self.validate(key, item)
            options = args[0]
        else:
            for key, item in args:
                self.validate(key, item)
        for key, item in options.items():
            self.validate(key, item)
        super().__init__(*args, **options)

    def validate(self, key: Any, item: Any) -> None:
        canstore, msg = self.validator(key, item)
        if not canstore:
            raise DiscriminationError(msg)

    def __setitem__(self, key: Any, item: Any) -> None:
        self.validate(key, item)
        b = key in self
        old = self.get(key)
        super().__setitem__(key, item)
        if b:
            self.__on_rm(key, old)
        self.__on_add(key, item)

    def __delitem__(self, key: Any) -> None:
        if key not in self:
            raise KeyError()
        v = self[key]
        super().__delitem__(key)
        self.__on_rm(key, v)

    def setdefault(self, key: Any, *args) -> Optional[Any]:
        default = args[0] if len(args) > 0 else None
        self.validate(key, default)
        b = key not in self
        r = super().setdefault(key, default)
        if b:
            self.__on_add(key, default)
        return r

    def update(self, other: dict, **kwargs) -> None:
        other = DiscriminatingDictionary(self.validator, other | kwargs)
        for k, v in other.items():
            self[k] = v

    def pop(self, key: Any) -> None:
        v = super().pop(key)
        self.__on_rm(key, v)

    def popitem(self) -> None:
        k, v = super().popitem()
        self.__on_rm(k, v)

    def clear(self) -> None:
        c = self.copy()
        super().clear()
        for k, v in c.items():
            self.__on_rm(k, v)

    def __on_add(self, key: Any, item: Any) -> None:
        if self.on_add is not None:
            self.on_add(key, item)

    def __on_rm(self, key: Any, item: Any) -> None:
        if self.on_remove is not None:
            self.on_remove(key, item)


class DiscriminationError(Exception):
    pass



Coordinate = tuple[int, int, int]


def block_is_in_box(block: Coordinate, box: tuple[Coordinate, Coordinate]) -> bool:
    """
    Returns True id the block is in the box
    block is (x, y, z)
    box is ((x0, y0, z0), (x1, y1, z1))
    """
    x, y, z = block
    xs = [box[0][0], box[1][0]]
    xs.sort()
    ys = [box[0][1], box[1][1]]
    ys.sort()
    zs = [box[0][1], box[1][1]]
    zs.sort()
    x_min, x_max = xs
    y_min, y_max = ys
    z_min, z_max = zs
    return x_min <= x <= x_max and y_min <= y <= y_max and z_min <= z <= z_max


def box_is_in_box(box1: tuple[Coordinate, Coordinate], box2: tuple[Coordinate, Coordinate]) -> bool:
    return block_is_in_box(box1[0], box2) and block_is_in_box(box1[1], box2)



from math import ceil, log
from time import time

import nbtlib
import numpy as np
from nbtlib.tag import Short, Byte, Int, Long, Double, String, List, Compound, ByteArray, IntArray
from typing_extensions import deprecated

from typing import Any, Generator, Callable, Optional

from .deprecation import deprecated_name
from .info import *
from .minecraft import BlockState, Entity, TileEntity, RequiredKeyMissingException
from .storage import LitematicaBitArray, DiscriminatingDictionary


class Schematic:
    """
    Represents a schematic file in the Litematic format.
    """

    name: str
    author: str
    description: str
    region: dict[str, 'Region']
    lm_version: int
    lm_subversion: int
    mc_version: int
    created: int
    modified: int
    __regions: DiscriminatingDictionary
    __preview: IntArray

    def __init__(self,
                 name: str = DEFAULT_NAME, author: str = "", description: str = "",
                 regions: Optional[dict[str, 'Region']] = None,
                 lm_version: int = LITEMATIC_VERSION, lm_subversion: int = LITEMATIC_SUBVERSION,
                 mc_version: int = MC_DATA_VERSION
                 ) -> None:
        """
        Schematic can be created by optionally providing metadata and regions, or leaving them blank or default.

        :param name:        The name of the schematic to write in the metadata
        :param author:      The name of the author to write in the metadata
        :param description: The description to write in the metadata
        :param regions:     Regions to populate the schematic with
        :param lm_version:  The litematic version (you are unlikely to ever need to use this)
        :param mc_version:  The Minecraft data version (you are unlikely to ever need to use this)
        """
        if regions is None:
            regions = {}
        self.author = author
        self.description = description
        self.name = name
        self.created = round(time() * 1000)
        self.modified = round(time() * 1000)
        self.__regions = DiscriminatingDictionary(self._can_add_region,
                                                  onadd=self.__on_region_add, onremove=self.__on_region_remove)
        self.__compute_enclosure()
        if regions is not None and len(regions) > 0:
            self.__regions.update(regions)
        self.mc_version = mc_version
        self.lm_version = lm_version
        self.lm_subversion = lm_subversion
        self.__preview = IntArray([])

    def save(self, file_path: str, update_meta: bool = True, save_soft: bool = True, gzipped: bool = True,
             byteorder: str = 'big') -> None:
        """
        Save this schematic to a file.

        :param file_path:   the filesystem path the schematic should be saved to
        :param update_meta: whether to update the schematic's metadata before saving
                            (see :func:`~litemapy.Schematic.update_metadata`)
        :param save_soft:   whether to add an entry to the metadata indicating the schematic was created with Litemapy
        :param gzipped:     whether to compress the NBT content with gzip (this is the normal behavior)
        :param byteorder:   endianness of NBT numbers (either "little" or "big", default is "big")

        :raises ValueError: if this schematic does not have any region
        """
        if update_meta:
            self.update_metadata()
        f = nbtlib.File(self.to_nbt(save_soft=save_soft), gzipped=gzipped, byteorder=byteorder)
        f.save(file_path)

    def to_nbt(self, save_soft: bool = True) -> Compound:
        """
        Write the schematic to an NBT tag.

        :param save_soft:   whether to add an entry to the metadata indicating the schematic was created with Litemapy

        :rtype: ~nbtlib.tag.Compound

        :raises ValueError: if this schematic does not have any region
        """
        if len(self.__regions) < 1:
            raise ValueError("Empty schematic does not have any regions")
        root = Compound()
        root["Version"] = Int(self.lm_version)
        root["SubVersion"] = Int(self.lm_subversion)
        root["MinecraftDataVersion"] = Int(self.mc_version)
        meta = Compound()
        enclose = Compound()
        enclose["x"] = Int(self.width)
        enclose["y"] = Int(self.height)
        enclose["z"] = Int(self.length)
        meta["EnclosingSize"] = enclose
        meta["Author"] = String(self.author)
        meta["Description"] = String(self.description)
        meta["Name"] = String(self.name)
        if save_soft:
            meta["Software"] = String(LITEMAPY_NAME + "_" + LITEMAPY_VERSION)
        meta["RegionCount"] = Int(len(self.regions))
        meta["TimeCreated"] = Long(self.created)
        meta["TimeModified"] = Long(self.modified)
        meta["TotalBlocks"] = Int(sum([reg.getblockcount() for reg in self.regions.values()]))
        meta["TotalVolume"] = Int(sum([reg.getvolume() for reg in self.regions.values()]))
        meta['PreviewImageData'] = self.__preview
        root["Metadata"] = meta
        regs = Compound()
        for name, region in self.regions.items():
            regs[name] = region.to_nbt()
        root["Regions"] = regs
        return root

    @deprecated_name("fromnbt")
    @staticmethod
    def from_nbt(nbt: Compound) -> 'Schematic':
        """
        Read a schematic from an NBT tag.

        :param nbt: a schematic serialized as an NBT tag

        :rtype:     Schematic

        :raises CorruptedSchematicError: if the schematic tag is malformed
        """
        meta: Compound = nbt["Metadata"]
        lm_version: Int = nbt["Version"]
        lm_subversion: Int = nbt.get("SubVersion", 0)
        mc_version: Int = nbt["MinecraftDataVersion"]
        width = int(meta["EnclosingSize"]["x"])
        height = int(meta["EnclosingSize"]["y"])
        length = int(meta["EnclosingSize"]["z"])
        author = str(meta["Author"])
        name = str(meta["Name"])
        desc = str(meta["Description"])
        regions: dict[str, 'Region'] = {}
        for key, value in nbt["Regions"].items():
            reg = Region.from_nbt(value)
            regions[str(key)] = reg
        schematic = Schematic(name=name, author=author, description=desc, regions=regions,
                              lm_version=lm_version, lm_subversion=lm_subversion,
                              mc_version=mc_version)
        if schematic.width != width:
            raise CorruptedSchematicError(
                "Invalid schematic width in metadata, excepted {} was {}".format(schematic.width, width))
        if schematic.height != height:
            raise CorruptedSchematicError(
                "Invalid schematic height in metadata, excepted {} was {}".format(schematic.height, height))
        if schematic.length != length:
            raise CorruptedSchematicError(
                "Invalid schematic length in metadata, excepted {} was {}".format(schematic.length, length))
        schematic.created = int(meta["TimeCreated"])
        schematic.modified = int(meta["TimeModified"])
        if "RegionCount" in meta and len(schematic.regions) != meta["RegionCount"]:
            raise CorruptedSchematicError("Number of regions in metadata does not match the number of parsed regions")
        if 'PreviewImageData' in meta.keys():
            schematic.__preview = meta['PreviewImageData']
        return schematic

    @deprecated_name("updatemeta")
    def update_metadata(self) -> None:
        """
        Update this schematic's metadata (set the modified time to the current time).
        """
        self.modified = round(time() * 1000)

    @staticmethod
    def load(file_path) -> 'Schematic':
        """
        Read a schematic from a file.

        :param file_path:   the filesystem path to the file to load

        :rtype:             Schematic

        :raises CorruptedSchematicError: if the schematic file is malformed in any way
        """
        nbt = nbtlib.File.load(file_path, True)
        return Schematic.from_nbt(nbt)

    def _can_add_region(self, name: str, region: 'Region') -> tuple[bool, str]:
        if type(name) != str:
            return False, "Region name should be a string"
        return True, ""

    def __on_region_add(self, name: str, region: 'Region') -> None:
        if self.__x_min is None:
            self.__x_min = region.min_schem_x()
        else:
            self.__x_min = min(self.__x_min, region.min_schem_x())
        if self.__x_max is None:
            self.__x_max = region.max_schem_x()
        else:
            self.__x_max = max(self.__x_max, region.max_schem_x())
        if self.__y_min is None:
            self.__y_min = region.min_schem_y()
        else:
            self.__y_min = min(self.__y_min, region.min_schem_y())
        if self.__y_max is None:
            self.__y_max = region.max_schem_y()
        else:
            self.__y_max = max(self.__y_max, region.max_schem_y())
        if self.__z_min is None:
            self.__z_min = region.min_schem_z()
        else:
            self.__z_min = min(self.__z_min, region.min_schem_z())
        if self.__z_max is None:
            self.__z_max = region.max_schem_z()
        else:
            self.__z_max = max(self.__z_max, region.max_schem_z())

    def __on_region_remove(self, name, region) -> None:
        bounding_box_changed: bool = self.__x_min == region.minschemx()
        bounding_box_changed = bounding_box_changed or self.__x_max == region.maxschemx()
        bounding_box_changed = bounding_box_changed or self.__y_min == region.minschemy()
        bounding_box_changed = bounding_box_changed or self.__y_max == region.maxschemy()
        bounding_box_changed = bounding_box_changed or self.__z_min == region.minschemz()
        bounding_box_changed = bounding_box_changed or self.__z_max == region.maxschemz()
        if bounding_box_changed:
            self.__compute_enclosure()

    def __compute_enclosure(self):
        x_min, x_max, y_min, y_max, z_min, z_max = None, None, None, None, None, None
        for region in self.__regions.values():
            x_min = min(x_min, region.minschemx()) if x_min is not None else region.minschemx()
            x_max = max(x_max, region.maxschemx()) if x_max is not None else region.maxschemx()
            y_min = min(y_min, region.minschemy()) if y_min is not None else region.minschemy()
            y_max = max(y_max, region.maxschemy()) if y_max is not None else region.maxschemy()
            z_min = min(z_min, region.minschemz()) if z_min is not None else region.minschemz()
            z_max = max(z_max, region.maxschemz()) if z_max is not None else region.maxschemz()
        self.__x_min = x_min
        self.__x_max = x_max
        self.__y_min = y_min
        self.__y_max = y_max
        self.__z_min = z_min
        self.__z_max = z_max

    @property
    def regions(self) -> DiscriminatingDictionary:
        """
        The regions in this schematic, as a dictionary.
        This is a read only property, and it is not possible to replace this dictionary.
        It can however be edited, as long as the suitable types are used.
        Using an incorrect type will raise a :class:`~litemapy.storage.DiscriminationError`.
        """
        return self.__regions

    @property
    def width(self) -> int:
        """
        The width of this Schematic's bounding box.
        See :ref:`Coordinate systems <coordinates>`.
        This property is read-only.
        """
        if self.__x_min is None or self.__x_max is None:
            return 0
        return self.__x_max - self.__x_min + 1

    @property
    def height(self) -> int:
        """
        The height of this Schematic's bounding box.
        See :ref:`Coordinate systems <coordinates>`.
        This property is read-only.
        """
        if self.__y_min is None or self.__y_max is None:
            return 0
        return self.__y_max - self.__y_min + 1

    @property
    def length(self) -> int:
        """
        The length of this Schematic's bounding box.
        See :ref:`Coordinate systems <coordinates>`.
        This property is read-only.
        """
        if self.__z_min is None or self.__z_max is None:
            return 0
        return self.__z_max - self.__z_min + 1

    @property
    def preview(self) -> IntArray:
        # TODO This is not documented on purpose because ideally we would make it return a usable Pillow Image object.
        return self.__preview

    @preview.setter
    def preview(self, value) -> None:
        self.__preview = value


class Region:
    """
    Represents a schematic region.
    """
    __x: int
    __y: int
    __z: int
    __width: int
    __height: int
    __length: int
    __palette: list[BlockState]
    __blocks: np.ndarray[np.uint32, Any]  # TODO replace any with the right shape when numpy supports its
    __entities: list[Entity]
    __block_ticks: list[Compound]
    __fluid_ticks: list[Compound]
    __tile_entities: list[TileEntity]

    def __init__(self, x, y, z, width, height, length) -> None:
        """
        :param x:       the X coordinate of the region in the schematic
        :param y:       the Y coordinate of the region in the schematic
        :param z:       the Z coordinate of the region in the schematic
        :param width:   the size of the region along the x-axis (can be negative!)
        :param height:  the size of the region along the y-axis (can be negative!)
        :param length:  the size of the region along the z-axis (can be negative!)

        :raises ValueError: if either width, height or length is 0
        """
        if width == 0 or height == 0 or length == 0:
            raise ValueError("Region dimensions cannot be 0")
        self.__x, self.__y, self.__z = x, y, z
        self.__width, self.__height, self.__length = width, height, length
        self.__palette = [AIR, ]
        self.__blocks = np.zeros((abs(width), abs(height), abs(length)), dtype=np.uint32)
        self.__entities = []
        self.__tile_entities = []
        self.__block_ticks = []
        self.__fluid_ticks = []

    def to_nbt(self) -> Compound:
        """
        Write this region to an NBT tag.

        """

        self._optimize_palette()

        root = Compound()
        pos = Compound()
        pos["x"] = Int(self.__x)
        pos["y"] = Int(self.__y)
        pos["z"] = Int(self.__z)
        root["Position"] = pos
        size = Compound()
        size["x"] = Int(self.__width)
        size["y"] = Int(self.__height)
        size["z"] = Int(self.__length)
        root["Size"] = size

        plt = List[Compound]([blk.to_nbt() for blk in self.__palette])
        root["BlockStatePalette"] = plt

        entities = List[Compound]([entity.to_nbt() for entity in self.__entities])
        root["Entities"] = entities

        tile_entities = List[Compound]([tile_entity.to_nbt() for tile_entity in self.__tile_entities])
        root["TileEntities"] = tile_entities

        root["PendingBlockTicks"] = List[Compound](self.__block_ticks)
        root["PendingFluidTicks"] = List[Compound](self.__fluid_ticks)

        arr = LitematicaBitArray(self.volume(), self.__get_needed_nbits())
        for x in range(abs(self.__width)):
            for y in range(abs(self.__height)):
                for z in range(abs(self.__length)):
                    ind = (y * abs(self.__width * self.__length)) + z * abs(self.__width) + x
                    arr[ind] = int(self.__blocks[x, y, z])
        root["BlockStates"] = arr._to_nbt_long_array()

        return root

    def to_sponge_nbt(self, mc_version: int = MC_DATA_VERSION, gzipped: bool = True,
                      endianness: str = 'big') -> nbtlib.nbt.File:
        """
        Returns the Region as an NBT Compound file that conforms to the Sponge Schematic Format (version 2) used by mods
        like WorldEdit.
        Check `the file format specification <https://github.com/SpongePowered/Schematic-Specification>`_
        for more information.

        :param mc_version:  Minecraft data version that is being emulated
                            (https://minecraft.wiki/w/Data_version).
                            Should not be critical for newer versions of Minecraft.
        :param gzipped:     Whether the NBT Compound file should be compressed
                            (WorldEdit only works with gzipped files).
        :param endianness:  Endianness of the resulting NBT Compound file
                            ('big' or 'little', WorldEdit only works with big endian files).

        :returns:           The Region represented as a Sponge Schematic NBT Compound file.
        """

        self._optimize_palette()

        # TODO Needs unit tests

        nbt = nbtlib.File(gzipped=gzipped, byteorder=endianness)

        nbt['DataVersion'] = Int(mc_version)
        nbt['Version'] = Int(SPONGE_VERSION)

        nbt['Width'] = Short(abs(self.__width))
        nbt['Height'] = Short(abs(self.__height))
        nbt['Length'] = Short(abs(self.__length))

        nbt['Offset'] = IntArray([Int(0), Int(0), Int(0)])  # not strictly necessary

        # process entities
        size = (self.__width, self.__height, self.__length)
        entities = List[Compound]()
        for entity in self.__entities:
            entity_tag = Compound()
            for key, value in entity.data.items():
                entity_tag[key] = value

            entity_tag['Pos'] = List[Double](
                [Double(coord - (0 if dim > 0 else (dim + 1))) for coord, dim in zip(entity.position, size)])
            keys = entity.data.keys()
            if 'TileX' in keys:
                entity_tag['TileX'] = Int(entity_tag['Pos'][0])
                entity_tag['TileY'] = Int(entity_tag['Pos'][1])
                entity_tag['TileZ'] = Int(entity_tag['Pos'][2])

            entity_tag['Id'] = entity_tag['id']
            del entity_tag['id']
            entities.append(entity_tag)

        nbt['Entities'] = entities

        # process tile entities
        tile_entities = List[Compound]()
        for tile_entity in self.__tile_entities:
            tile_entity_tag = Compound()
            for key, value in tile_entity.data.items():
                tile_entity_tag[key] = value

            tile_entity_tag['Pos'] = IntArray([Int(coord) for coord in tile_entity.position])
            for key in ['x', 'y', 'z']:
                del tile_entity_tag[key]
            tile_entities.append(tile_entity_tag)

        nbt['BlockEntities'] = tile_entities

        # process block palette
        nbt['PaletteMax'] = Int(len(self.__palette))
        palette = Compound()
        for i, block in enumerate(self.__palette):
            state = block.to_block_state_identifier()
            palette[state] = Int(i)

        nbt['Palette'] = palette

        # process blocks
        block_array = []
        for i in range(abs(self.__width * self.__height * self.__length)):
            blocks_per_layer = abs(self.__width * self.__length)
            y = int(i / blocks_per_layer)
            i_in_layer = i % blocks_per_layer
            z = int(i_in_layer / abs(self.__width))
            x = i_in_layer % abs(self.__width)
            block_array.append(self.__blocks[x, y, z])

        nbt['BlockData'] = ByteArray([Byte(id) for id in block_array])

        return nbt

    @staticmethod
    def from_sponge_nbt(nbt: Compound) -> tuple['Region', int]:
        """
        Returns a Litematica Region based on an NBT Compound that conforms to the Sponge Schematic Format (version 2)
        used by mods like WorldEdit.
        Check `the file format specification <https://github.com/SpongePowered/Schematic-Specification>`_
        for more information.

        :param nbt: The Sponge schematic NBT Compound.

        :returns:   a Litematica Region built from the Sponge schematic
                    and the Minecraft data version that the Sponge schematic was created for.
        """

        # TODO Needs unit tests

        mc_version = nbt['DataVersion']
        width = int(nbt['Width'])
        height = int(nbt['Height'])
        length = int(nbt['Length'])
        region = Region(0, 0, 0, width, height, length)
        offset = nbt['Offset']

        # process entities
        for entity in nbt['Entities']:
            if 'Id' not in entity.keys():
                raise RequiredKeyMissingException('Id')
            entity['id'] = entity['Id']
            del entity['Id']

            ent = Entity(entity)
            position = [coord - off for coord, off in zip(ent.position, offset)]
            ent.position = (position[0], position[1], position[2])
            region.entities.append(ent)

        # process tile entities
        tile_entities = nbt['BlockEntities']
        for tile_entity in tile_entities:
            if 'Id' not in tile_entity.keys():
                raise RequiredKeyMissingException('Id')
            tile_entity['id'] = tile_entity['Id']
            del tile_entity['Id']

            tent = TileEntity.from_nbt(tile_entity)
            tent.position = tent.data['Pos']
            del tile_entity['Pos']
            region.tile_entities.append(tent)

        # process blocks and let __setitem__() automatically generate the palette
        palette = nbt['Palette']
        palette_dict = {}
        for block, index in palette.items():
            property_dict = {}
            if block.find('[') == -1:
                block_id = block
            else:
                entries = block.split('[')
                block_id = entries[0]
                properties = entries[1].replace(']', '').split(',')
                for property in properties:
                    key, value = property.split('=')
                    property_dict[key] = value

            block_state = BlockState(block_id, **property_dict)
            palette_dict[int(index)] = block_state

        for i, index in enumerate(nbt['BlockData']):
            blocks_per_layer = width * length
            y = int(i / blocks_per_layer)
            i_in_layer = i % blocks_per_layer
            z = int(i_in_layer / width)
            x = i_in_layer % width
            region[x, y, z] = palette_dict[int(index)]

        return region, mc_version

    def to_structure_nbt(self, mc_version=MC_DATA_VERSION, gzipped=True, byteorder='big') -> nbtlib.nbt.File:
        """
        Returns the Region as an NBT Compound file that conforms to Minecraft's structure NBT files.

        :param mc_version:  Minecraft data version that is being emulated
                            (https://minecraft.wiki/w/Data_version).
                            Should not be critical for newer versions of Minecraft.
        :param gzipped:     Whether the NBT Compound file should be compressed
                            (Vanilla Minecraft only works with gzipped files).
        :param byteorder:   Endianness of the resulting NBT Compound file
                            ('big' or 'little', Vanilla Minecraft only works with big endian files).

        :returns:           The Region represented as a Minecraft structure NBT file.
        """

        # TODO Needs unit tests

        self._optimize_palette()

        structure = nbtlib.File(gzipped=gzipped, byteorder=byteorder)

        structure['size'] = List[Int]([abs(self.__width), abs(self.__height), abs(self.__length)])
        structure['DataVersion'] = Int(mc_version)

        # process entities
        size = (self.__width, self.__height, self.__length)
        entities = List[Compound]()
        for entity in self.__entities:
            entity_tag = Compound()
            entity_tag['nbt'] = entity.data
            entity_tag['pos'] = List[Double](
                [Double(coord - (0 if dim > 0 else (dim + 1))) for coord, dim in zip(entity.position, size)])
            entity_tag['blockPos'] = List[Int](
                [Int(coord - (0 if dim > 0 else (dim + 1))) for coord, dim in zip(entity.position, size)])
            entities.append(entity_tag)

        structure['entities'] = entities

        # create tile entity dictionary to add them correctly to the block list later
        tile_entity_dict = {}
        for tile_entity in self.__tile_entities:
            tile_entity_tag = Compound()
            for key, value in tile_entity.data.items():
                if key not in ['x', 'y', 'z']:
                    tile_entity_tag[key] = value

            tile_entity_dict[tile_entity.position] = tile_entity_tag

        # process palette
        structure['palette'] = List[Compound]([block.to_nbt() for block in self.__palette])

        # process blocks
        blocks = List[Compound]()
        for x, y, z in np.ndindex(self.__blocks.shape):
            block = Compound()
            position = (x, y, z)
            if position in tile_entity_dict.keys():
                block['nbt'] = tile_entity_dict[position]
            block['pos'] = List[Int]([Int(coord) for coord in position])
            block['state'] = Int(self.__blocks[x, y, z])
            blocks.append(block)

        structure['blocks'] = blocks

        return structure

    @staticmethod
    def from_structure_nbt(structure: Compound) -> tuple['Region', str]:
        """
        Returns a Litematica Region based on an NBT Compound that conforms to Minecraft's structure NBT files.

        :param structure:   The Minecraft structure NBT Compound.

        :returns:           A Litematica Region built from the Minecraft structure
                            and the Minecraft data version that the structure was created for
        """

        # TODO Needs unit tests

        mc_version = structure['DataVersion']
        size = structure['size']
        width = int(size[0])
        height = int(size[1])
        length = int(size[2])
        region = Region(0, 0, 0, width, height, length)

        # process entities
        for entity in structure['entities']:
            ent = Entity(entity['nbt'])
            ent.position = entity['pos']
            region.entities.append(ent)

        # process blocks and let __setitem__() automatically generate the palette
        palette = structure['palette']
        for block in structure['blocks']:
            x, y, z = block['pos']
            state = block['state']
            region[x, y, z] = BlockState.from_nbt(palette[state])
            if 'nbt' in block.keys():
                tile_entity = TileEntity(block['nbt'])
                tile_entity.position = block['pos']
                region.tile_entities.append(tile_entity)

        return region, mc_version

    def __getitem__(self, position: tuple[int, int, int]) -> BlockState:
        x, y, z = self.__region_coordinates_to_store_coordinates(*position)
        return self.__palette[self.__blocks[x, y, z]]

    @deprecated("Region.getblock() is deprecated. Use array style syntax instead: region[x, y, z]")
    def getblock(self, x: int, y: int, z: int) -> BlockState:
        return self.__getitem__((x, y, z))

    def __setitem__(self, position: tuple[int, int, int], block: BlockState) -> None:
        x, y, z = self.__region_coordinates_to_store_coordinates(*position)
        if block in self.__palette:
            i = self.__palette.index(block)
        else:
            self.__palette.append(block)
            i = len(self.__palette) - 1
        self.__blocks[x, y, z] = i

    @deprecated("Region.setblock() is deprecated. Use array style syntax instead: region[x, y, z]")
    def setblock(self, x: int, y: int, z: int, block: BlockState):
        return self.__setitem__((x, y, z), block)

    def __contains__(self, block: BlockState) -> bool:
        return block in self.__palette and self.__palette.index(block) in self.__blocks

    @deprecated_name("getblockcount")
    def count_blocks(self) -> int:
        """
        Counts the number of blocks in the region.

        :returns: the number of non-air blocks in the region
        """

        # air is index zero
        return np.count_nonzero(self.__blocks)

    def __region_coordinates_to_store_coordinates(self, x: int, y: int, z: int) -> tuple[int, int, int]:
        if self.__width < 0:
            x -= self.__width + 1
        if self.__height < 0:
            y -= self.__height + 1
        if self.__length < 0:
            z -= self.__length + 1
        return x, y, z

    @deprecated_name("getvolume")
    def volume(self) -> int:
        """
        Computes this region's volume.

        :returns: this region volume in blocks
        """
        return abs(self.__width * self.__height * self.__length)

    def __get_needed_nbits(self) -> int:
        return max(ceil(log(len(self.__palette), 2)), 2)

    @deprecated_name("fromnbt")
    @staticmethod
    def from_nbt(nbt: Compound) -> 'Region':
        """
        Read a region from an NBT tag.

        :param nbt: an NBT tag to read the region from
        """
        pos = nbt["Position"]
        x = int(pos["x"])
        y = int(pos["y"])
        z = int(pos["z"])
        size = nbt["Size"]
        width = int(size["x"])
        height = int(size["y"])
        length = int(size["z"])
        region = Region(x, y, z, width, height, length)
        del region.__palette[0]
        for block_nbt in nbt["BlockStatePalette"]:
            block = BlockState.from_nbt(block_nbt)
            region.__palette.append(block)

        for entity_nbt in nbt["Entities"]:
            entity = Entity.from_nbt(entity_nbt)
            region.entities.append(entity)

        for tile_entity_nbt in nbt["TileEntities"]:
            block = TileEntity.from_nbt(tile_entity_nbt)
            region.tile_entities.append(block)

        blocks = nbt["BlockStates"]
        nbits = region.__get_needed_nbits()
        bit_array = LitematicaBitArray.from_nbt_long_array(blocks, region.volume(), nbits)
        for x in range(abs(width)):
            for y in range(abs(height)):
                for z in range(abs(length)):
                    ind = (y * abs(width * length)) + z * abs(width) + x
                    region.__blocks[x][y][z] = bit_array[ind]

        for block_ticks in nbt["PendingBlockTicks"]:
            region.__block_ticks.append(block_ticks)

        for fluid_ticks in nbt["PendingFluidTicks"]:
            region.__fluid_ticks.append(fluid_ticks)

        return region

    @deprecated_name("minschemx")
    def min_schem_x(self) -> int:
        """
        :returns:   the minimum X coordinate of this region in the schematics coordinate system
        """
        return min(self.__x, self.__x + self.width + 1)

    @deprecated_name("maxschemx")
    def max_schem_x(self) -> int:
        """
        :returns:   the maximum X coordinate of this region in the schematics coordinate system
        """
        return max(self.__x, self.__x + self.width - 1)

    @deprecated_name("minschemy")
    def min_schem_y(self) -> int:
        """
        :returns:   the minimum Y coordinate of this region in the schematics coordinate system
        """
        return min(self.__y, self.__y + self.height + 1)

    @deprecated_name("maxschemy")
    def max_schem_y(self) -> int:
        """
        :returns:   the maximum Y coordinate of this region in the schematics coordinate system
        """
        return max(self.__y, self.__y + self.height - 1)

    @deprecated_name("minschemz")
    def min_schem_z(self) -> int:
        """
        :returns:   the minimum Z coordinate of this region in the schematics coordinate system
        """
        return min(self.__z, self.__z + self.length + 1)

    @deprecated_name("maxschemz")
    def max_schem_z(self) -> int:
        """
        :returns:   the maximum Z coordinate of this region in the schematics coordinate system
        """
        return max(self.__z, self.__z + self.length - 1)

    @deprecated_name("minx")
    def min_x(self) -> int:
        """
        :returns:   the minimum X coordinate of this region in its own coordinate system
        """
        return min(0, self.width + 1)

    @deprecated_name("maxx")
    def max_x(self) -> int:
        """
        :returns:   the maximum X coordinate of this region in its own coordinate system
        """
        return max(0, self.width - 1)

    @deprecated_name("miny")
    def min_y(self) -> int:
        """
        :returns:   the minimum Y coordinate of this region in its own coordinate system
        """
        return min(0, self.height + 1)

    @deprecated_name("maxy")
    def max_y(self) -> int:
        """
        :returns:   the maximum Y coordinate of this region in its own coordinate system
        """
        return max(0, self.height - 1)

    @deprecated_name("minz")
    def min_z(self) -> int:
        """
        :returns:   the minimum Z coordinate of this region in its own coordinate system
        """
        return min(0, self.length + 1)

    @deprecated_name("maxz")
    def max_z(self) -> int:
        """
        :returns:   the maximum Z coordinate of this region in its own coordinate system
        """
        return max(0, self.length - 1)

    @deprecated_name("xrange")
    def range_x(self) -> range:
        """
        :returns:   the range of coordinates this region contains along its X axis
        """
        return range(self.min_x(), self.max_x() + 1)

    @deprecated_name("yrange")
    def range_y(self) -> range:
        """
        :returns:   the range of coordinates this region contains along its Y axis
        """
        return range(self.min_y(), self.max_y() + 1)

    @deprecated_name("zrange")
    def range_z(self) -> range:
        """
        :returns:   the range of coordinates this region contains along its Z axis
        """
        return range(self.min_z(), self.max_z() + 1)

    @deprecated_name("allblockpos")
    def block_positions(self) -> Generator[tuple[int, int, int], None, None]:
        """
        :returns:   an iterator over the coordinates this region contains in its own coordinate system
        """
        for x in self.range_x():
            for y in self.range_y():
                for z in self.range_z():
                    yield x, y, z

    @property
    def x(self) -> int:
        """
        The X coordinate of the region within the schematic's coordinate system.
        This property is read only.
        """
        return self.__x

    @property
    def y(self) -> int:
        """
        The Y coordinate of the region within the schematic's coordinate system.
        This property is read only.
        """
        return self.__y

    @property
    def z(self) -> int:
        """
        The Z coordinate of the region within the schematic's coordinate system.
        The property is read only.
        """
        return self.__z

    @property
    def width(self) -> int:
        """
        The width of the region.
        This property is read only.
        """
        return self.__width

    @property
    def height(self) -> int:
        """
        The height of the region.
        This property is read only.
        """
        return self.__height

    @property
    def length(self) -> int:
        """
        The length of the region.
        This property is read only.
        """
        return self.__length

    @property
    def entities(self) -> list[Entity]:
        """
        The entities within the region.
        """
        return self.__entities

    @property
    def tile_entities(self) -> list[TileEntity]:
        """
        The tile entities within the region.
        """
        return self.__tile_entities

    @property
    def block_ticks(self) -> list[Compound]:
        # TODO We are not exporting the documentation for this because it still exposes the raw NBT data
        return self.__block_ticks

    @property
    def fluid_ticks(self) -> list[Compound]:
        # TODO We are not exporting the documentation for this because it still exposes the raw NBT data
        return self.__fluid_ticks

    @property
    def palette(self) -> tuple[BlockState, ...]:
        """
        The palette used to store blocks within the region.
        Each entry in the palette is assured to be unique.
        Expected the first palette entry which is always "minecraft:air",
        each entry is assured to have at least one instance in the region.
        """
        self._optimize_palette()
        return tuple(self.__palette)

    def as_schematic(self, name: str = DEFAULT_NAME, author: str = "", description: str = "",
                     mc_version: int = MC_DATA_VERSION) -> Schematic:
        """
        Creates a schematic that contains that region at the origin.

        :param name:        a name for both the region and the schematic
        :param author:      an author for the schematic
        :param description: a description for the schematic
        :param mc_version:  The Minecraft data version (you are unlikely to ever need to use this)
        """
        return Schematic(name=name, author=author, description=description, regions={name: self}, mc_version=mc_version)

    def __replace_palette_index(self, old_index: int, new_index: int) -> None:
        if old_index == new_index:
            return
        self.__blocks[self.__blocks == old_index] = new_index

    def _optimize_palette(self) -> None:
        # Functions that work directly with the palette like filter or replace
        # may introduce duplicates or unused entries in the palette.
        # For this reason, it is necessary to clean things up before exporting
        # block content in any way
        new_palette = []
        for old_index, state in enumerate(self.__palette):
            # Skip unused entries, except air that needs to remain at index 0
            if old_index != 0 and old_index not in self.__blocks:
                continue
            # Do not copy duplicate entries multiple times
            for i, other_state in enumerate(new_palette):
                if state == other_state:
                    new_index = i
                    break
            else:
                # Keep that entry
                new_index = len(new_palette)
                new_palette.append(state)
            # Update blocks to reflect the new palette
            self.__replace_palette_index(old_index, new_index)
        self.__palette = new_palette

    def filter(self, function: Callable[[BlockState], BlockState]) -> None:
        """
        Replaces all occurrences of :class:`BlockState` with others by providing a mapping function.
        This method works with the palette directly and the mapping function is therefore only called
        once per block state type in the region, and not for every position.
        This is a lot faster than manually iterating over region coordinates.

        :param function: a mapping function
        """
        self.__palette = list(map(function, self.__palette))

        # We need to ensure we always have air at palette index 0
        if self.__palette[0] != AIR:
            self.__palette.append(self.__palette[0])
            self.__replace_palette_index(0, len(self.__palette) - 1)
            self.__palette[0] = AIR

    def replace(self, replace: BlockState, replace_with: BlockState) -> None:
        """
        Replace all occurrences of a :class:`BlockState` with another.
        This method works with the palette directly, in O(1) time,
        and is therefore a lot more efficient than looping over the content.

        :param replace:         the blockstate to replace
        :param replace_with:    a new blockstate to replace the old one with
        """
        try:
            index = self.__palette.index(replace)
        except ValueError:
            return  # Nothing to do
        if index == 0:
            # We are replacing air, that's not good
            self.__palette.append(replace_with)
            self.__replace_palette_index(0, len(self.__palette) - 1)
        else:
            self.__palette[index] = replace_with


AIR = BlockState("minecraft:air")


class CorruptedSchematicError(Exception):
    pass



from json import dumps
from nbtlib.tag import Int, Double, String, List, Compound, Base
from typing_extensions import deprecated

from .deprecation import deprecated_name
from .storage import DiscriminatingDictionary

from typing import Any, Optional, Union


EntityPosition = tuple[float, float, float]
EntityRotation = tuple[float, float]
EntityMotion = tuple[float, float, float]

BlockPosition = tuple[int, int, int]


class BlockState:
    """
    Represents an in-game block.
    :class:`BlockState` are immutable.
    """

    __block_id: str
    __properties: DiscriminatingDictionary
    __identifier_cache: Optional[str]

    def __init__(self, block_id: str, **properties: str) -> None:
        """
        A block state has a block ID and a dictionary of properties.

        :param block_id:    the identifier of the block (e.g. *minecraft:stone*)
        :param properties:  the properties of the block state as keyword parameters (e.g. *facing="north"*)
        """
        self.__block_id = assert_valid_identifier(block_id)
        self.__properties = DiscriminatingDictionary(self.__validate, properties)
        self.__identifier_cache = None

    def to_nbt(self) -> Compound:
        """
        Writes this block state to an nbt tag.
        """
        root = Compound()
        root["Name"] = String(self.id)
        properties: dict[str, str] = {String(k): String(v) for k, v in self.__properties.items()}
        if len(properties) > 0:
            root["Properties"] = Compound(properties)
        return root

    @deprecated_name("fromnbt")
    @staticmethod
    def from_nbt(nbt: Compound) -> 'BlockState':
        """
        Reads a :class:`BlockState` from an nbt tag.
        """
        block_id = assert_valid_identifier(str(nbt["Name"]))
        if "Properties" in nbt:
            properties: dict[str, str] = {str(k): str(v) for k, v in nbt["Properties"].items()}
        else:
            properties: dict[str, str] = {}
        block = BlockState(block_id, **properties)
        return block

    @property
    def id(self) -> str:
        """
        The block's identifier.
        """
        return self.__block_id

    @property
    @deprecated("Use BlockState.id instead")
    def blockid(self) -> str:
        return self.__block_id

    @deprecated_name("with_blockid")
    def with_id(self, block_id: str) -> 'BlockState':
        """
        Returns a new :class:`BlockState` with the same properties as this one but a different block id.

        :param block_id:  the block id for the new :class:`BlockState`
        """
        assert_valid_identifier(block_id)
        return BlockState(block_id, **self.__properties)

    def with_properties(self, **properties: Optional[str]) -> 'BlockState':
        """
        Returns a new copy of this :class:`BlockState` with new values for the properties given in keyword arguments.
        Using `None` as a property value removes it.

        :param properties:  the new properties as keyword arguments

        :returns: A copy of this :class:`BlockState` with the given properties updated to new values
        """
        none_properties = list(map(lambda kv: kv[0], filter(lambda kv: kv[1] is None, properties.items())))
        other = BlockState(self.id)
        other.__properties.update(self.__properties)
        for prop_name in none_properties:
            other.__properties.pop(prop_name)
            properties.pop(prop_name)
        other.__properties.update(properties)
        return other

    def __validate(self, k: Any, v: Any) -> tuple[bool, str]:
        if type(k) is not str or type(v) is not str:
            return False, "BlockState properties should be a string => string dictionary"
        return True, ""

    def to_block_state_identifier(self, skip_empty: bool = True) -> str:
        """
        Returns an identifier that represents the BlockState in the Sponge Schematic Format (version 2).
        Format: block_type[properties]
        Example: minecraft:oak_sign[rotation=0,waterlogged=false]

        :param skip_empty:  Whether empty brackets should be excluded if the BlockState has no properties.

        :returns: An identifier that represents the BlockState in a Sponge schematic.
        """

        if skip_empty and self.__identifier_cache is not None:
            # The result is cached when skip_empty is True,
            # but this function is used to implement __hash__,
            # so we can use functools.cache
            return self.__identifier_cache

        # TODO Needs unit tests

        identifier: str = self.__block_id
        if not skip_empty or len(self.__properties) > 0:
            state = dumps(self.__properties, separators=(',', '='), sort_keys=True)
            state = state.replace('{', '[').replace('}', ']')
            state = state.replace('"', '').replace("'", '')
            identifier += state

        if skip_empty:
            self.__identifier_cache = identifier

        return identifier

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BlockState):
            return False
        return other.__block_id == self.__block_id and other.__properties == self.__properties

    def __hash__(self) -> int:
        return hash(self.to_block_state_identifier())

    def __repr__(self) -> str:
        return self.to_block_state_identifier(skip_empty=True)

    def __getitem__(self, key: str) -> Optional[str]:
        return self.__properties[key]

    def __len__(self) -> int:
        return len(self.__properties)


class Entity:
    """
    A Minecraft entity.
    Each entity is identified by a type identifier (e.g. minecraft:skeleton)
    and has a position within a region, as well as a rotation and a velocity vector.
    Most also have arbitrary data depending on their type
    (e.g. a sheep has a tag for its color and one indicating whether it has been sheared).
    """

    _id: str
    _data: Compound
    _position: EntityPosition
    _rotation: EntityRotation
    _motion: EntityMotion

    # TODO Needs unit tests

    def __init__(self, str_or_nbt: Union[str, Compound]) -> None:
        # TODO Refactor to only have a from_nbt static method instead of allowing nbt into the constructor
        """
        :param str_or_nbt:  either the entity identifier as a string, in which case all other tag will be default,
                            or an bnt compound tag with the entitie's data.
        """

        if isinstance(str_or_nbt, str):
            self._data = Compound({'id': String(str_or_nbt)})
        else:
            self._data = str_or_nbt

        keys = self._data.keys()
        if 'id' not in keys:
            raise RequiredKeyMissingException('id')
        if 'Pos' not in keys:
            self._data['Pos'] = List[Double]([Double(0.), Double(0.), Double(0.)])
        if 'Rotation' not in keys:
            self._data['Rotation'] = List[Double]([Double(0.), Double(0.)])
        if 'Motion' not in keys:
            self._data['Motion'] = List[Double]([Double(0.), Double(0.), Double(0.)])

        self._id = assert_valid_identifier(self._data['id'])
        position = [float(coord) for coord in self._data['Pos']]
        self._position = (position[0], position[1], position[2])
        rotation = [float(coord) for coord in self._data['Rotation']]
        self._rotation = (rotation[0], rotation[1])
        motion = [float(coord) for coord in self._data['Motion']]
        self._motion = (motion[0], motion[1], motion[2])

    def to_nbt(self) -> Compound:
        """
        Save this entity as an NBT tag.
        """
        return self._data

    @deprecated_name("fromnbt")
    @staticmethod
    def from_nbt(nbt: Compound) -> 'Entity':
        """
        Read an entity from an nbt tag.

        :param nbt: An NBT tag with the entity's data
        """
        return Entity(nbt)

    def add_tag(self, key: str, tag) -> None:
        self._data[key] = tag
        if key == 'id':
            self._id = str(tag)
        if key == 'Pos':
            position = [float(coord) for coord in tag]
            self._position = (position[0], position[1], position[2])
        if key == 'Rotation':
            rotation = [float(coord) for coord in tag]
            self._rotation = (rotation[0], rotation[1])
        if key == 'Motion':
            motion = [float(coord) for coord in tag]
            self._motion = (motion[0], motion[1], motion[2])

    def get_tag(self, key: str) -> Base:
        try:
            return self._data[key]
        except KeyError:
            raise

    @property
    def data(self) -> Compound:
        # TODO Not documented because it exposes NBT
        return self._data

    @data.setter
    def data(self, data: Compound) -> None:
        # TODO Not documented because it exposes NBT
        self._data = Entity(data).data
        self._id = str(self._data['id'])
        position = [float(coord) for coord in self._data['Pos']]
        self._position = (position[0], position[1], position[2])
        rotation = [float(coord) for coord in self._data['Rotation']]
        self._rotation = (rotation[0], rotation[1])
        motion = [float(coord) for coord in self._data['Motion']]
        self._motion = (motion[0], motion[1], motion[2])

    @property
    def id(self) -> str:
        """
        This entity's type identifier (e.g. *minecraft:pig* )
        """
        return self._id

    @id.setter
    def id(self, id: str) -> None:
        self._id = id
        self._data['id'] = String(self._id)

    @property
    def position(self) -> EntityPosition:
        """
        The position of the entity.
        """
        return self._position

    @position.setter
    def position(self, position: EntityPosition) -> None:
        self._position = position
        self._data['Pos'] = List[Double]([Double(coord) for coord in self._position])

    @property
    def rotation(self) -> EntityRotation:
        """
        The rotation of the entity.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, rotation: EntityRotation) -> None:
        self._rotation = rotation
        self._data['Rotation'] = List[Double]([Double(coord) for coord in self._rotation])

    @property
    def motion(self) -> EntityMotion:
        """
        The velocity vector of the entity.
        """
        return self._motion

    @motion.setter
    def motion(self, motion: EntityMotion) -> None:
        self._motion = motion
        self._data['Motion'] = List[Double]([Double(coord) for coord in self._motion])


class TileEntity:
    # TODO Needs unit tests
    """
    A tile entity, also often referred to as block entities,
    is a type of entity which complements a block state to store additional data
    (e.g. containers like chest both have a block state that stores properties
    like their id ( *minecraft:chest* ) and orientation, and tile entity that stores their content.
    For this reason, the :class:`TileEntity` class does not store an ID  but only a position.
    The ID can be inferred by looking up the :class:`BlockState` as the same position in the :class:`Region`.
    """
    _data: Compound
    _position: BlockPosition

    def __init__(self, nbt: Compound) -> None:
        # TODO Not documented because it only exposes NBT
        self._data = nbt
        keys = self._data.keys()

        if 'x' not in keys:
            self._data['x'] = Int(0)
        if 'y' not in keys:
            self._data['y'] = Int(0)
        if 'z' not in keys:
            self._data['z'] = Int(0)
        position = [int(self._data[coord]) for coord in ['x', 'y', 'z']]
        self._position = (position[0], position[1], position[2])

    def to_nbt(self) -> Compound:
        """
        Saves the tile entity to NBT tag.
        """
        return self._data

    @deprecated_name("fromnbt")
    @staticmethod
    def from_nbt(nbt: Compound) -> 'TileEntity':
        """
        Reads a tile entity from an NBT tag.

        :param nbt: the tile entity's data as an NBT tag
        """
        return TileEntity(nbt)

    def add_tag(self, key: str, tag) -> None:
        # TODO Not documented because it exposes NBT
        self._data[key] = tag

        pos: tuple[int, int, int] = self._position
        if key == 'x':
            self._position = (int(tag), pos[1], pos[2])
        if key == 'y':
            self._position = (pos[0], int(tag), pos[2])
        if key == 'z':
            self._position = (pos[0], pos[1], int(tag))

    def get_tag(self, key: str) -> Base:
        # TODO Not documented because it exposes NBT
        try:
            return self._data[key]
        except KeyError:
            raise

    @property
    def data(self) -> Compound:
        # TODO Not documented because it exposes NBT
        return self._data

    @data.setter
    def data(self, data: Compound):
        self._data = TileEntity(data).data
        position = [int(self._data[coord]) for coord in ['x', 'y', 'z']]
        self._position = (position[0], position[1], position[2])

    @property
    def position(self) -> BlockPosition:
        """
        The tile entity's position within the :class:`Region`/
        """
        return self._position

    @position.setter
    def position(self, position: BlockPosition):
        self._position = position
        for coord, index in [('x', 0), ('y', 1), ('z', 2)]:
            self._data[coord] = Int(self._position[index])


def is_valid_identifier(identifier: str) -> bool:
    """
    Checks if a string is a valid identifier (aka. ResourceLocation in Mojmap).
    """
    # Check taken from Minecraft 1.20.1 ResourceLocation
    allowed_chars = "_-abcdefghijklmnopqrstuvwxyz0123456789.:"
    separator = False
    for char in identifier:
        if char not in allowed_chars:
            return False
        if char == ":":
            # Now parsing the path part
            separator = True
            allowed_chars = "_-abcdefghijklmnopqrstuvwxyz0123456789./"
    return separator


class InvalidIdentifier(ValueError):
    identifier: str

    def __init__(self, identifier: str) -> None:
        super().__init__(f'Invalid identifier "{identifier}"')


def assert_valid_identifier(identifier: str) -> str:
    """
    Checks whether a string is a valid identifier (aka  ResourceLocation in Mojmap),
    and raises InvalidIdentifierError if it is not.
    The name "identifier" is from Yarn mappings but makes more sens in this context.

    :returns: the identifier
    :raises CorruptedSchematicError: if provided string is not a valid identifier
    """

    if not is_valid_identifier(identifier):
        raise InvalidIdentifier(identifier)
    return identifier


class RequiredKeyMissingException(Exception):

    def __init__(self, key: str, message: str = 'The required key is missing in the (Tile)Entity\'s NBT Compound'):
        self.key = key
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.key} -> {self.message}'



from litemapy import Schematic, Region, BlockState
from os import walk
from constants import *
import helper
from tempfile import TemporaryDirectory

AIR = BlockState("minecraft:air")

valid_files = []
for directory, child_directory, file_names in walk(VALID_LITEMATIC_DIRECTORY):
    for file_name in file_names:
        valid_files.append(path.join(directory, file_name))


def test_valid_litematics_do_not_raise_exception_when_loaded():
    for file_path in valid_files:
        Schematic.load(file_path)


def test_regions_have_accurate_sizes():
    sch = Schematic()
    assert sch.width == 0
    assert sch.height == 0
    assert sch.length == 0
    reg1 = Region(0, 0, 0, 10, 10, 10)
    reg2 = Region(90, 0, 0, 10, 10, 10)
    reg3 = Region(50, 0, 0, 10, 10, 10)
    reg4 = Region(49, 0, 0, -10, 10, 10)
    sch.regions["a"] = reg1
    assert sch.width == 10
    assert sch.height == 10
    assert sch.length == 10
    sch.regions["b"] = reg2
    assert sch.width == 100
    sch.regions["c"] = reg3
    assert sch.width == 100
    del sch.regions["b"]
    assert sch.width == 60
    sch.regions["c"] = reg4
    assert sch.width == 50


def test_region_min_and_max_are_accurate():
    reg = Region(0, 0, 0, 10, 20, 30)
    assert reg.min_schem_x() == 0
    assert reg.max_schem_x() == 9
    assert reg.min_schem_y() == 0
    assert reg.max_schem_y() == 19
    assert reg.min_schem_z() == 0
    assert reg.max_schem_z() == 29
    assert reg.min_x() == 0
    assert reg.max_x() == 9
    assert reg.min_y() == 0
    assert reg.max_y() == 19
    assert reg.min_z() == 0
    assert reg.max_z() == 29
    reg = Region(0, 0, 0, -10, -20, -30)
    assert reg.min_schem_x() == -9
    assert reg.max_schem_x() == 0
    assert reg.min_schem_y() == -19
    assert reg.max_schem_y() == 0
    assert reg.min_schem_z() == -29
    assert reg.max_schem_z() == 0
    assert reg.min_x() == -9
    assert reg.max_x() == 0
    assert reg.min_y() == -19
    assert reg.max_y() == 0
    assert reg.min_z() == -29
    assert reg.max_z() == 0
    reg = Region(10, 10, 10, 10, 10, 10)
    assert reg.min_schem_x() == 10
    assert reg.max_schem_x() == 19
    assert reg.min_schem_y() == 10
    assert reg.max_schem_y() == 19
    assert reg.min_schem_z() == 10
    assert reg.max_schem_z() == 19
    assert reg.min_x() == 0
    assert reg.max_x() == 9
    assert reg.min_y() == 0
    assert reg.max_y() == 9
    assert reg.min_z() == 0
    assert reg.max_z() == 9
    reg = Region(-10, -10, -10, 10, 10, 10)
    assert reg.min_schem_x() == -10
    assert reg.max_schem_x() == -1
    assert reg.min_schem_y() == -10
    assert reg.max_schem_y() == -1
    assert reg.min_schem_z() == -10
    assert reg.max_schem_z() == -1
    assert reg.min_x() == 0
    assert reg.max_x() == 9
    assert reg.min_y() == 0
    assert reg.max_y() == 9
    assert reg.min_z() == 0
    assert reg.max_z() == 9
    reg = Region(-10, -10, -10, -10, -10, -10)
    assert reg.min_schem_x() == -19
    assert reg.max_schem_x() == -10
    assert reg.min_schem_y() == -19
    assert reg.max_schem_y() == -10
    assert reg.min_schem_z() == -19
    assert reg.max_schem_z() == -10
    assert reg.min_x() == -9
    assert reg.max_x() == 0
    assert reg.min_y() == -9
    assert reg.max_y() == 0
    assert reg.min_z() == -9
    assert reg.max_z() == 0


def test_are_random_schematics_preserved_when_reading_and_writing():
    temporary_directory = TemporaryDirectory()
    for i in range(10):
        write_schematic = helper.randomschematic()
        file_path = path.join(temporary_directory.name, write_schematic.name + ".litematic")
        write_schematic.save(file_path)
        read_schematic = Schematic.load(file_path)

        # Assert metadata is equals
        assert write_schematic.name == read_schematic.name
        assert write_schematic.author == read_schematic.author
        assert write_schematic.description == read_schematic.description
        assert write_schematic.width == read_schematic.width
        assert write_schematic.height == read_schematic.height
        assert write_schematic.length == read_schematic.length
        assert len(write_schematic.regions) == len(read_schematic.regions)
        for name, write_region in write_schematic.regions.items():
            read_region = read_schematic.regions[name]

            # Assert computed values are equal
            assert write_region.min_x() == read_region.min_x()
            assert write_region.max_x() == read_region.max_x()
            assert write_region.min_y() == read_region.min_y()
            assert write_region.max_y() == read_region.max_y()
            assert write_region.min_z() == read_region.min_z()
            assert write_region.max_z() == read_region.max_z()
            assert write_region.min_schem_x() == read_region.min_schem_x()
            assert write_region.max_schem_x() == read_region.max_schem_x()
            assert write_region.min_schem_y() == read_region.min_schem_y()
            assert write_region.max_schem_y() == read_region.max_schem_y()
            assert write_region.min_schem_z() == read_region.min_schem_z()
            assert write_region.max_schem_z() == read_region.max_schem_z()

            # Assert all blocks are equal
            for x, y, z in write_region.allblockpos():
                ws = write_region[x, y, z]
                rs = read_region[x, y, z]
                assert ws == rs

            assert_valid_palette(write_region)

    temporary_directory.cleanup()


def test_region_filter():
    def do_filter(before_schematic, after_schematic, function):
        print(f"Comparing litematic files {before_schematic} and {after_schematic}")
        before_schematic = path.join(FILTER_LITEMATIC_DIRECTORY, before_schematic)
        after_schematic = path.join(FILTER_LITEMATIC_DIRECTORY, after_schematic)
        before_schematic = Schematic.load(before_schematic)
        after_schematic = Schematic.load(after_schematic)
        assert len(before_schematic.regions) == 1, "Invalid test litematic"
        assert len(after_schematic.regions) == 1, "Invalid test litematic"
        (before_schematic,) = before_schematic.regions.values()
        (after_schematic,) = after_schematic.regions.values()
        assert before_schematic.width == after_schematic.width, "Invalid test litematic"
        assert before_schematic.height == after_schematic.height, "Invalid test litematic"
        assert before_schematic.length == after_schematic.length, "Invalid test litematic"
        before_schematic.filter(function)
        for x in before_schematic.range_x():
            for y in before_schematic.range_y():
                for z in before_schematic.range_z():
                    state_1 = before_schematic[x, y, z]
                    state_2 = after_schematic[x, y, z]
                    assert state_1 == state_2
        assert_valid_palette(before_schematic)

    def all_blue_filter(b: BlockState):
        return BlockState("minecraft:light_blue_concrete")

    do_filter('rainbow-line.litematic', 'blue-line.litematic', all_blue_filter)

    red = BlockState("minecraft:red_concrete")
    blue = BlockState("minecraft:blue_concrete")

    def black_red_white_blue(b: BlockState):
        if b.id == "minecraft:black_concrete":
            return red
        if b.id == "minecraft:white_concrete":
            return blue
        return b

    do_filter('black-white.litematic', 'red-blue.litematic', black_red_white_blue)

    def glassify(state: BlockState):
        if "water" in state.id:
            return BlockState("minecraft:blue_stained_glass")
        elif state.id == "minecraft:sand":
            return BlockState('minecraft:yellow_stained_glass')
        elif state.id == "minecraft:dirt":
            return BlockState("minecraft:brown_stained_glass")
        elif state.id == "minecraft:stone":
            return BlockState("minecraft:light_gray_stained_glass")
        elif state.id in ("minecraft:grass_block", "minecraft:birch_leaves"):
            return BlockState("minecraft:green_stained_glass")
        elif state.id == "minecraft:birch_log":
            return BlockState("minecraft:white_stained_glass")
        elif state.id == "minecraft:copper_ore":
            return BlockState("minecraft:orange_stained_glass")
        elif state.id == "minecraft:grass":
            return BlockState("minecraft:green_stained_glass_pane", east="true", north="true", south="true",
                              west="true", waterlogged="false")
        return state

    do_filter('tree.litematic', 'tree-glass.litematic', glassify)

    def wool_to_concrete(b: BlockState):
        return b.with_id(b.id.replace('wool', 'concrete'))

    do_filter('concrete-wool.litematic', 'concrete-full.litematic', wool_to_concrete)


def assert_valid_palette(region: Region):
    palette = region.palette

    assert palette[0] == AIR, "Palette does not have air at index 0"

    entries = set()
    for entry in palette:
        assert entry not in entries, f"Palette has duplicate entry: {entry}"
        entries.add(entry)

    blocks = {region[p] for p in region.block_positions()}
    for entry in palette:
        if entry == AIR:
            continue
        assert entry in blocks, f"Palette contains an unused entry"


def test_unused_palette_entries_get_pruned():
    region = Region(0, 0, 0, 10, 10, 10)
    region[0, 0, 0] = BlockState("minecraft:stone")
    region[0, 0, 0] = AIR
    assert_valid_palette(region)


def test_region_getitem_setitem():
    region = Region(0, 0, 0, 1, 1, 1)
    stone = BlockState("minecraft:stone")
    region[0, 0, 0] = stone
    assert region[0, 0, 0] == stone


def test_region_in():
    region = Region(0, 0, 0, 10, 10, 10)
    stone = BlockState("minecraft:stone")
    assert stone not in region
    region[0, 0, 0] = stone
    assert stone in region
    region[0, 0, 0] = AIR
    assert stone not in region


def test_replace():
    region = Region(0, 0, 0, 10, 10, 10)
    stone = BlockState("minecraft:stone")
    log = BlockState("minecraft:oak_log")
    grass = BlockState("minecraft:grass_block")
    region[1, 0, 0] = stone
    region[2, 0, 0] = log
    assert region[0, 0, 0] == AIR
    region.replace(AIR, grass)
    assert region[0, 0, 0] == grass
    assert region.palette[0] == AIR
    region.replace(log, stone)
    assert region[2, 0, 0] == stone


def test_subversion():
    schematic = Schematic.load(path.join(VALID_LITEMATIC_DIRECTORY, "Subversion.litematic"))
    assert schematic.lm_subversion == 1
    with TemporaryDirectory() as temp:
        schematic.lm_subversion = 1337
        name = path.join(temp, "Subversion.litematic")
        schematic.save(name)
        schematic = Schematic.load(name)
    assert schematic.lm_subversion == 1337



from typing import Callable


def deprecated_name(old_name: str):
    class Wrapper:
        original: Callable

        def __init__(self, original: Callable):
            self.original = original

        def __set_name__(self, owner, name):
            setattr(owner, name, self.original)
            setattr(owner, old_name, self.original)

    return Wrapper



import litemapy.boxes as boxes

origin = (0, 0, 0)
block1 = (-2, -2, -2)
block2 = (2, 2, 2)
block3 = (1, 2, 3)
block4 = (-4, -4, -4)
block5 = (4, 4, 4)
centered_box_1 = (block1, block2)
centered_box_2 = (block4, block5)
box1 = (origin, block2)
box2 = (block1, block5)
box3 = (block1, block4)


def test_block_is_in_box():
    assert boxes.block_is_in_box(origin, centered_box_1)
    assert not (boxes.block_is_in_box(block3, (block1, block2)))


def test_box_is_in_box():
    assert boxes.box_is_in_box(centered_box_1, centered_box_2)
    assert boxes.box_is_in_box(box1, centered_box_2)
    assert not boxes.box_is_in_box(centered_box_2, centered_box_1)
    assert not boxes.box_is_in_box(box1, box3)
